"""
Measure Optics
--------------

This module contains high-level functions to manage most functionality of ``optics_measurements``.
It provides functions to compute various lattice optics parameters from frequency spectra.
"""
import datetime
import os
import sys
from collections import OrderedDict
from copy import deepcopy

import numpy as np
import pandas as pd
import tfs

from omc3 import __version__ as VERSION
from omc3.definitions.constants import PLANES
from omc3.optics_measurements import (beta_from_amplitude, beta_from_phase,
                                      chromatic, dispersion, dpp, iforest,
                                      interaction_point, kick, phase, rdt,
                                      tune, crdt, coupling)
from omc3.optics_measurements.constants import (CHROM_BETA_NAME, ERR, EXT)
from omc3.utils import iotools, logging_tools

LOGGER = logging_tools.get_logger(__name__, level_console=logging_tools.INFO)
LOG_FILE = "measure_optics.log"


def measure_optics(input_files, measure_input):
    """
    Main function to compute various lattice optics parameters from frequency spectra.

    Args:
        input_files: `InputFiles` object containing frequency spectra files (linx/y).
        measure_input: `OpticsInput` object containing analysis settings.

    Returns:
    """
    LOGGER.info(f"Calculating optics parameters - code version {VERSION}")
    iotools.create_dirs(measure_input.outputdir)
    logging_tools.add_module_handler(logging_tools.file_handler(
        os.path.join(measure_input.outputdir, LOG_FILE)))
    tune_dict = tune.calculate(measure_input, input_files)
    common_header = _get_header(measure_input, tune_dict)
    invariants = {}
    phase_dict = {}
    for plane in PLANES:
        phase_dict[plane], out_dfs  = phase.calculate(measure_input, input_files, tune_dict, plane)
        phase.write(out_dfs, [common_header]*4, measure_input.outputdir, plane)
        phase.write_special(measure_input, phase_dict[plane]['free'], tune_dict[plane]["QF"], plane)
        if measure_input.only_coupling:
            continue
        beta_df, beta_header = beta_from_phase.calculate(measure_input, tune_dict, phase_dict[plane], common_header, plane)
        beta_from_phase.write(beta_df, beta_header, measure_input.outputdir, plane)

        ratio = beta_from_amplitude.calculate(measure_input, input_files, tune_dict, beta_df, common_header, plane)
        invariants[plane] = kick.calculate(measure_input, input_files, ratio, common_header, plane)
        ip_df = interaction_point.betastar_from_phase(measure_input, phase_dict[plane]['free'])
        interaction_point.write(ip_df, common_header, measure_input.outputdir, plane)
        dispersion.calculate_orbit(measure_input, input_files, common_header, plane)
        dispersion.calculate_dispersion(measure_input, input_files, common_header, plane)
        if plane == "X":
            dispersion.calculate_normalised_dispersion(measure_input, input_files, beta_df, common_header)
    coupling.calculate_coupling(measure_input, input_files, phase_dict, tune_dict, common_header)
    if measure_input.only_coupling:
        return
    if 'rdt' in measure_input.nonlinear:
        iotools.create_dirs(os.path.join(measure_input.outputdir, "rdt"))
        rdt.calculate(measure_input, input_files, tune_dict, phase_dict, invariants, common_header)
    if 'crdt' in measure_input.nonlinear:
        iotools.create_dirs(os.path.join(measure_input.outputdir, "crdt"))
        crdt.calculate(measure_input, input_files, invariants, common_header)
    if measure_input.chromatic_beating:
        chromatic_beating(input_files, measure_input, tune_dict)


def chromatic_beating(input_files, measure_input, tune_dict):
    """
    Main function to compute chromatic optics beating.

    Args:
        tune_dict:
        input_files: `InputFiles` object containing frequency spectra files (linx/y).
        measure_input:` OpticsInput` object containing analysis settings.

    Returns:
    """
    dpps = np.array([dpp_val for dpp_val in set(input_files.dpps("X"))])
    if np.max(dpps) - np.min(dpps) == 0.0:
        return
    for plane in PLANES:
        betas = []
        for dpp_val in dpps:
            dpp_meas_input = deepcopy(measure_input)
            dpp_meas_input["dpp"] = dpp_val
            phase_dict, out_dfs = phase.calculate(dpp_meas_input, input_files, tune_dict, plane)
            beta_df, _ = beta_from_phase.calculate(dpp_meas_input, tune_dict, phase_dict, OrderedDict(), plane)
            betas.append(beta_df)
        output_df = chromatic.calculate_w_and_phi(betas, dpps, input_files, measure_input, plane)
        tfs.write(os.path.join(measure_input.outputdir, f"{CHROM_BETA_NAME}{plane.lower()}{EXT}"), output_df, {}, save_index="NAME")


def _get_header(meas_input, tune_dict):
    compensation = {'model': "by model", 'equation': "by equation", 'none': "None"}
    return OrderedDict([('Measure_optics:version', VERSION),
                        ('Command', f"{sys.executable} {' '.join(sys.argv)}"),
                        ('CWD', os.getcwd()),
                        ('Date', datetime.datetime.today().strftime("%d. %B %Y, %H:%M:%S")),
                        ('Model_directory', meas_input.accelerator.model_dir),
                        ('Compensation', compensation[meas_input.compensation]),
                        ('Q1', tune_dict["X"]["QF"]),
                        ('Q2', tune_dict["Y"]["QF"])])


class InputFiles(dict):
    """
    Stores the input files, provides methods to gather quantity specific data

    Public methods:
        - ``get_dpps`` (plane)
        - ``get_joined_frame`` (plane, columns, zero_dpp=False, how='inner')
        - ``get_columns`` (frame, column)
        - ``get_data`` (frame, column)
    """
    def __init__(self, files_to_analyse, optics_opt):
        super(InputFiles, self).__init__(zip(PLANES, ([], [])))
        read_files = isinstance(files_to_analyse[0], str)
        for file_in in files_to_analyse:
            for plane in PLANES:
                df_to_load = (tfs.read(f"{file_in}.lin{plane.lower()}").set_index("NAME", drop=False)
                              if read_files else file_in[plane])
                df_to_load.index.name = None
                self[plane].append(self._repair_backwards_compatible_frame(df_to_load, plane))

        if len(self['X']) + len(self['Y']) == 0:
            raise IOError("No valid input files")
        dpp_values = dpp.calculate_dpoverp(self, optics_opt)
        LOGGER.info(f"DPPS: {dpp_values}")
        amp_dpp_values = dpp.calculate_amp_dpoverp(self, optics_opt)
        LOGGER.info(f"DPP_AMPS: {amp_dpp_values}")
        for plane in PLANES:
            if optics_opt.isolation_forest:
                self[plane] = iforest.clean_with_isolation_forest(self[plane], optics_opt, plane)
            self[plane] = dpp.append_dpp(self[plane], dpp.arrange_dpps(dpp_values))
            self[plane] = dpp.append_amp_dpp(self[plane], amp_dpp_values)

    @staticmethod  # TODO later remove
    def _repair_backwards_compatible_frame(df, plane):
        """
        Multiplies unscaled amplitudes by 2 to get from complex amplitudes to the real ones.
        This is for backwards compatibility with Drive.
        """
        df[f"AMP{plane}"] = df.loc[:, f"AMP{plane}"].to_numpy() * 2
        if f"NATAMP{plane}" in df.columns:
            df[f"NATAMP{plane}"] = df.loc[:, f"NATAMP{plane}"].to_numpy() * 2
        return df

    def dpps(self, plane: str) -> np.ndarray:
        """
        Gathers measured DPPs from input files corresponding to given plane

        Args:
            plane: marking the horizontal or vertical plane, **X** or **Y**.

        Returns:
            A `np.ndarray` of DPPs.
        """
        return np.array([df.DPP for df in self[plane]])

    def dpp_frames(self, plane, dpp_value):
        dpp_dfs = []
        for i in np.argwhere(np.abs(self.dpps(plane) - dpp_value) < 1e-6).T[0]:
            dpp_dfs.append(self[plane][i])
        if len(dpp_dfs) == 0:
            raise ValueError(f"No data found for dp/p {dpp}")
        return dpp_dfs

    def _all_frames(self, plane):
        return self[plane]

    def joined_frame(self, plane, columns, dpp_value=None, dpp_amp=False, how='inner'):
        """
        Constructs merged DataFrame from InputFiles.

        Args:
            plane: marking the horizontal or vertical plane, **X** or **Y**.
            columns: list of columns from input files.
            dpp_value: merges only files with given ``dpp_value``.
            dpp_amp: merges only files with non-zero dpp amplitude (i.e. 3Dkicks).
            how: whi way to use for merging: ``inner`` (intersection) or ``outer`` (union),
                default is ``inner``.

        Returns:
            A merged `DataFrame` from `InputFiles`.
        """
        if how not in ['inner', 'outer']:
            raise RuntimeWarning("'how' should be either 'inner' or 'outer', 'inner' will be used.")
        frames_to_join = self.dpp_frames(plane, dpp_value) if dpp_value is not None else self._all_frames(plane)
        if dpp_amp:
            frames_to_join = [df for df in frames_to_join if df.DPPAMP > 0]
        if len(frames_to_join) == 0:
            raise ValueError(f"No data found for non-zero |dp/p|")
        joined_frame = pd.DataFrame(frames_to_join[0]).reindex(columns=columns, fill_value=np.nan)
        if len(frames_to_join) > 1:
            for i, df in enumerate(frames_to_join[1:]):
                joined_frame = pd.merge(joined_frame, df.reindex(columns=columns, fill_value=np.nan),
                                        how=how, left_index=True,
                                        right_index=True, suffixes=('', '__' + str(i + 1)))
        for column in columns:
            joined_frame.rename(columns={column: column + '__0'}, inplace=True)
        return joined_frame

    def bpms(self, plane=None, dpp_value=None):
        if plane is None:
            return self.bpms(plane="X", dpp_value=dpp_value).intersection(self.bpms(plane="Y", dpp_value=dpp_value))
        indices = [df.index for df in (self.dpp_frames(plane, dpp_value) if dpp_value is not None else self._all_frames(plane))]
        for ind in indices[1:]:
            indices[0] = indices[0].intersection(ind)
        return indices[0]

    def calibrate(self, calibs):
        if calibs is None:
            return
        for plane in PLANES:
            for i in range(len(self[plane])):
                data = pd.merge(self[plane][i].loc[:, ["AMP" + plane]], calibs[plane], how='left',
                                left_index=True, right_index=True).fillna(
                    value={"CALIBRATION": 1., "ERROR_CALIBRATION": 0.})
                self[plane][i][f"AMP{plane}"] = self[plane][i].loc[:, f"AMP{plane}"] * data.loc[:, "CALIBRATION"]
                self[plane][i][f"{ERR}AMP{plane}"] = data.loc[:, "ERROR_CALIBRATION"]  # TODO

    @ staticmethod
    def get_columns(frame, column):
        """
        Returns list of columns of frame corresponding to column in original files.

        Args:
            frame:  joined frame.
            column: name of column in original files.

        Returns:
            list of columns.
        """
        str_list = list(frame.columns[frame.columns.str.startswith(column + '__')].to_numpy())
        new_list = list(map(lambda s: s[len(f"{column}__"):], str_list))
        new_list.sort(key=int)
        return [f"{column}__{x}" for x in new_list]

    def get_data(self, frame, column) -> np.ndarray:
        """
        Returns data in columns of frame corresponding to column in original files.

        Args:
            frame:  joined frame.
            column: name of column in original files.

        Returns:
            A `np.narray` corresponding to column in original files.
        """
        columns = self.get_columns(frame, column)
        return frame.loc[:, columns].to_numpy()


def copy_calibration_files(outputdir, calibrationdir):
    if calibrationdir is None:
        return None
    calibs = {}
    for plane in PLANES:
        cal_file = f"calibration_{plane.lower()}.out"
        iotools.copy_item(os.path.join(calibrationdir, cal_file), os.path.join(outputdir, cal_file))
        calibs[plane] = tfs.read(os.path.join(outputdir, cal_file)).set_index("NAME")
    return calibs



"""
Constants
---------

Specific constants to be used in correction, to help with consistency.
"""
# Column Names -----------------------------------------------------------------
# Pre- and Suffixe
ERR = "ERR"  # Error of the measurement
RMS = "RMS"  # Root-Mean-Square
RES = "RES"  # Rescaled measurement
DELTA = "DELTA"  # Delta between measurement and model (sometimes beating)
MDL = "MDL"  # Model
VALUE = "VALUE"
WEIGHT = "WEIGHT"
ERROR = "ERROR"
MODEL = "MODEL"
DIFF = "DIFF"

# Names
NAME = "NAME"
NAME2 = "NAME2"
S = "S"
BETA = "BET"
BETABEAT = "BB"
DISP = "D"
F1001 = "F1001"
F1010 = "F1010"
INCR = "incr"
NORM_DISP = "ND"
PHASE = "PHASE"
PHASE_ADV = "MU"
TUNE = "Q"



r"""
Global Correction
-----------------

Iterative Correction Scheme.

The response matrices :math:`R_{O}` for the observables :math:`O` (e.g. BBX, PHASEX, ...)
are loaded from a file and then the equation

.. math:: R_{O} \cdot \delta var = O_{meas} - O_{model}
    :label: eq1

is being solved for :math:`\delta var` via a chosen method (at the moment only numpys pinv,
which creates a pseudo-inverse via svd is used).

The response matrices are hereby merged into one matrix for all observables to solve for all
:math:`\delta var` at the same time.

To normalize the observables to another ``weigths`` (W) can be applied.

Furthermore, an ``errorcut``, specifying the maximum errorbar for a BPM to be used, and
``modelcut``, specifying the maximum distance between measurement and model for a BPM to be used,
can be defined. Data from BPMs outside of those cut-values will be discarded.
These cuts are defined for each observable separately.

After each iteration the model variables are changed by :math:`-\delta var` and the
observables are recalculated by Mad-X.
:eq:`eq1` is then solved again.

Input arguments are split into correction arguments and accelerator arguments.
The former are listed below, the latter depend on the accelerator you want
to use. Check :ref:`modules/model:Model` to see which ones are needed.

**Arguments:**

*--Required--*

- **meas_dir**:

    Path to the directory containing the measurement files.


- **output_dir**:

    Path to the directory where to write the output files.


*--Optional--*

- **beta_file_name**:

    Prefix of the beta file to use. E.g.: getkmodbeta

    default: ``beta_phase_``


- **errorcut** *(float)*:

    Reject BPMs whose error bar is higher than the corresponding input.
    Input in order of optics_params.


- **fullresponse_path**:

    Path to the fullresponse binary file.If not given, calculates the
    response analytically.


- **max_iter** *(int)*:

    Maximum number of correction re-iterations to perform. A value of `0`
    means the correction is calculated once.

    default: ``3``


- **method** *(str)*:

    Optimization method to use.

    choices: ``('pinv', 'omp')``

    default: ``pinv``


- **min_corrector_strength** *(float)*:

    Minimum (absolute) strength of correctors.

    default: ``0.0``


- **modelcut** *(float)*:

    Reject BPMs whose deviation to the model is higher than the
    correspoding input. Input in order of optics_params.


- **n_correctors** *(int)*:

    Maximum number of correctors to use. (Method: 'omp')


- **optics_params** *(str)*:

    List of parameters to correct upon (e.g. BETX BETY)

    choices: ``('PHASEX', 'PHASEY', 'BBX', 'BBY', 'BETX', 'BETY', 'DX', 'DY', 'NDX', 'Q', 'F1001R', 'F1001I', 'F1010R', 'F1010I')``

    default: ``['PHASEX', 'PHASEY', 'BETX', 'BETY', 'NDX', 'Q']``


- **output_filename**:

    Identifier of the output files.

    default: ``changeparameters_iter``


- **svd_cut** *(float)*:

    Cutoff for small singular values of the pseudo inverse. (Method:
    'pinv')Singular values smaller than rcond*largest_singular_value are
    set to zero

    default: ``0.01``


- **update_response**:

    Update the (analytical) response per iteration.

    action: ``store_true``


- **use_errorbars**:

    Take into account the measured errorbars in the correction.

    action: ``store_true``


- **variable_categories**:

    List of names of the variables classes to use.

    default: ``['MQM', 'MQT', 'MQTL', 'MQY']``


- **weights** *(float)*:

    Weight to apply to each measured quantity. Input in order of
    optics_params.



Possible problems and notes (lmalina, 2020):
 * error-based weights default? likely - but be careful with low tune errors vs svd cut in pseudoinverse
 * manual creation of pd.DataFrame varslist, deltas? maybe tunes in tfs_pandas single value or a column?
 * There should be some summation/renaming for iterations
 * For two beam correction
 * The two beams can be treated separately until the calculation of correction
 * Values missing in the response (i.e. correctors of the other beam) shall be treated as zeros
 * Missing a part that treats the output from LSA

"""
from pathlib import Path
from typing import Dict

from generic_parser import DotDict
from generic_parser.entrypoint_parser import EntryPointParameters, entrypoint

from omc3.correction import handler
from omc3.correction.constants import (BETA, BETABEAT, DISP, F1001, F1010,
                                       NORM_DISP, PHASE, TUNE)
from omc3.model import manager
from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)

OPTICS_PARAMS_CHOICES = (f"{PHASE}X", f"{PHASE}Y",
                         f"{BETA}X", f"{BETA}Y",
                         f"{NORM_DISP}X",
                         f"{TUNE}",
                         f"{DISP}X", f"{DISP}Y",
                         f"{F1001}R", f"{F1001}I", f"{F1010}R", f"{F1010}I")

CORRECTION_DEFAULTS = {
    "optics_file": None,
    "output_filename": "changeparameters_iter",
    "svd_cut": 0.01,
    "optics_params": OPTICS_PARAMS_CHOICES[:6],
    "variable_categories": ["MQM", "MQT", "MQTL", "MQY"],
    "beta_file_name": "beta_phase_",
    "method": "pinv",
    "max_iter": 3,
}


def correction_params():
    params = EntryPointParameters()
    params.add_parameter(name="meas_dir",
                         required=True,
                         help="Path to the directory containing the measurement files.",)
    params.add_parameter(name="output_dir",
                         required=True,
                         help="Path to the directory where to write the output files.", )
    params.add_parameter(name="fullresponse_path",
                         help="Path to the fullresponse binary file.If not given, "
                              "calculates the response analytically.",)
    params.add_parameter(name="optics_params",
                         type=str,
                         nargs="+",
                         default=list(CORRECTION_DEFAULTS["optics_params"]),
                         choices=OPTICS_PARAMS_CHOICES,
                         help=f"List of parameters to correct upon (e.g. {BETA}X {BETA}Y)", )
    params.add_parameter(name="output_filename",
                         default=CORRECTION_DEFAULTS["output_filename"],
                         help="Identifier of the output files.", )
    params.add_parameter(name="min_corrector_strength",
                         type=float,
                         default=0.,
                         help="Minimum (absolute) strength of correctors.",)
    params.add_parameter(name="modelcut",
                         nargs="+",
                         type=float,
                         help="Reject BPMs whose deviation to the model is higher "
                              "than the correspoding input. Input in order of optics_params.",)
    params.add_parameter(name="errorcut",
                         nargs="+",
                         type=float,
                         help="Reject BPMs whose error bar is higher than the corresponding "
                              "input. Input in order of optics_params.",)
    params.add_parameter(name="weights",
                         nargs="+", type=float,
                         help="Weight to apply to each measured quantity. "
                              "Input in order of optics_params.",)
    params.add_parameter(name="variable_categories",
                         nargs="+",
                         default=CORRECTION_DEFAULTS["variable_categories"],
                         help="List of names of the variables classes to use.", )
    params.add_parameter(name="beta_file_name",
                         default=CORRECTION_DEFAULTS["beta_file_name"],
                         help="Prefix of the beta file to use. E.g.: getkmodbeta", )
    params.add_parameter(name="method",
                         type=str,
                         choices=("pinv", "omp"),
                         default=CORRECTION_DEFAULTS["method"],
                         help="Optimization method to use.", )
    params.add_parameter(name="svd_cut",
                         type=float,
                         default=CORRECTION_DEFAULTS["svd_cut"],
                         help="Cutoff for small singular values of the pseudo inverse. "
                              "(Method: 'pinv')Singular values smaller than "
                              "rcond*largest_singular_value are set to zero", )
    params.add_parameter(name="n_correctors",
                         type=int,
                         help="Maximum number of correctors to use. (Method: 'omp')")
    params.add_parameter(name="max_iter",
                         type=int,
                         default=CORRECTION_DEFAULTS["max_iter"],
                         help="Maximum number of correction re-iterations to perform. "
                              "A value of `0` means the correction is calculated once.", )
    params.add_parameter(name="use_errorbars",
                         action="store_true",
                         help="Take into account the measured errorbars in the correction.", )
    params.add_parameter(name="update_response",
                         action="store_true",
                         help="Update the (analytical) response per iteration.", )
    return params


@entrypoint(correction_params())
def global_correction_entrypoint(opt: DotDict, accel_opt) -> None:
    """Do the global correction. Iteratively."""
    LOG.info("Starting Iterative Global Correction.")
    opt = _check_opt_add_dicts(opt)
    opt = _add_hardcoded_paths(opt)
    opt.output_dir.mkdir(parents=True, exist_ok=True)
    accel_inst = manager.get_accelerator(accel_opt)
    handler.correct(accel_inst, opt)


def _check_opt_add_dicts(opt: dict) -> dict:  # acts inplace...
    """ Check on options and put in missing values """
    def_dict = _get_default_values()
    opt.optics_params = [param.replace(f"{BETABEAT}", f"{BETA}") for param in opt.optics_params]
    for key in ("modelcut", "errorcut", "weights"):
        if opt[key] is None:
            opt[key] = [def_dict[key][p] for p in opt.optics_params]
        elif len(opt[key]) != len(opt.optics_params):
            raise AttributeError(f"Length of {key} is not the same as of the optical parameters!")
        opt[key] = dict(zip(opt.optics_params, opt[key]))
    opt.meas_dir = Path(opt.meas_dir)
    opt.output_dir = Path(opt.output_dir)
    return opt


def _add_hardcoded_paths(opt: DotDict) -> DotDict:  # acts inplace...
    opt.change_params_path = opt.output_dir / f"{opt.output_filename}.madx"
    opt.change_params_correct_path = opt.output_dir / f"{opt.output_filename}_correct.madx"
    opt.knob_path = opt.output_dir / f"{opt.output_filename}.tfs"
    return opt


# Define functions here, to new optics params
def _get_default_values() -> Dict[str, Dict[str, float]]:
    return {
        "modelcut": {
            f"{PHASE}X": 0.05,
            f"{PHASE}Y": 0.05,
            f"{BETA}X": 0.2,
            f"{BETA}Y": 0.2,
            f"{DISP}X": 0.2,
            f"{DISP}Y": 0.2,
            f"{NORM_DISP}X": 0.2,
            f"{TUNE}": 0.1,
            f"{F1001}R": 0.2,
            f"{F1001}I": 0.2,
            f"{F1010}R": 0.2,
            f"{F1010}I": 0.2,
        },
        "errorcut": {
            f"{PHASE}X": 0.035,
            f"{PHASE}Y": 0.035,
            f"{BETA}X": 0.02,
            f"{BETA}Y": 0.02,
            f"{DISP}X": 0.02,
            f"{DISP}Y": 0.02,
            f"{NORM_DISP}X": 0.02,
            f"{TUNE}": 0.027,
            f"{F1001}R": 0.02,
            f"{F1001}I": 0.02,
            f"{F1010}R": 0.02,
            f"{F1010}I": 0.02,
        },
        "weights": {
            f"{PHASE}X": 1,
            f"{PHASE}Y": 1,
            f"{BETA}X": 0,
            f"{BETA}Y": 0,
            f"{DISP}X": 0,
            f"{DISP}Y": 0,
            f"{NORM_DISP}X": 0,
            f"{TUNE}": 10,
            f"{F1001}R": 0,
            f"{F1001}I": 0,
            f"{F1010}R": 0,
            f"{F1010}I": 0,
        },
    }


if __name__ == "__main__":
    with logging_tools.DebugMode(active=False, log_file="iterative_correction.log"):
        global_correction_entrypoint()



"""
Plot Spectrum - Waterfall Plotter
---------------------------------

Waterfall plotting functionality for spectrum plotter.
"""
import matplotlib
import numpy as np
from generic_parser import DotDict
from matplotlib import pyplot as plt, colors

from omc3.plotting.spectrum.utils import (plot_lines, PLANES, LABEL_Y_WATERFALL,
                                          LABEL_X, AMPS, FREQS, output_plot)
from omc3.plotting.utils.annotations import get_fontsize_as_float
from omc3.utils import logging_tools

LOG = logging_tools.getLogger(__name__)


def create_waterfall_plots(figures: dict, opt: DotDict) -> None:
    """ Main loop for waterfall plot creation. """
    LOG.debug(f"  ...creating Waterfall Plot")

    for fig_id, fig_cont in figures.items():
        LOG.debug(f'   Plotting Figure: {fig_id}.')
        fig_cont.fig.canvas.manager.set_window_title(fig_id)

        _plot_waterfall(fig_cont, opt.line_width, opt.cmap, opt.common_plane_colors)
        plot_lines(fig_cont, opt.lines)
        _format_axes(fig_cont, opt.limits, opt.ncol_legend)
        output_plot(fig_cont)

    if opt.show:
        plt.show()


def _plot_waterfall(fig_cont, line_width, cmap, common_plane_colors):
    """ Create the waterfall plot for this figure container. """
    for idx_plane, plane in enumerate(PLANES):
        ax = fig_cont.axes[idx_plane]
        norm = _get_color_norm(fig_cont.minmax, plane, common_plane_colors)
        for idx_data, (label, data) in enumerate(fig_cont.data.items()):
            if data[plane] is None:
                continue

            freqs = data[plane][FREQS]
            amps = data[plane][AMPS]

            if line_width == "auto":
                _plot_color_mesh(ax, freqs, amps, idx_data, cmap, norm)
            else:
                _plot_vlines(ax, freqs, amps, idx_data, cmap, norm, line_width)
        ax.figure.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax)


def _get_color_norm(minmax, plane, common):
    """ Returns the color-norm calculated from the min and max values. """
    if not common:
        return colors.LogNorm(*minmax[plane])

    return colors.LogNorm(min(mm[0] for mm in minmax.values()),
                          max(mm[1] for mm in minmax.values()))


def _plot_color_mesh(ax, freqs, amps, idx_data, cmap, norm):
    """ Plots the frequencies as a mesh, with amplitudes as colors. """
    freqs = freqs.sort_values()
    amps = amps.loc[freqs.index]

    freqs_mesh = np.tile(np.array([*freqs.to_numpy().T, .5]), [2, 1])
    y_mesh = np.tile([idx_data - 0.5, idx_data + 0.5], [len(freqs) + 1, 1]).T
    ax.pcolormesh(freqs_mesh, y_mesh, amps.to_frame().T,
                  cmap=cmap, norm=norm, zorder=-3)


def _plot_vlines(ax, freqs, amps, idx_data, cmap, norm, line_width):
    """ Plots the frequencies as vertical lines, with amplitudes as colors. """
    lines = ax.vlines(x=freqs, ymin=idx_data - .5, ymax=idx_data + .5,
                      linestyles='solid', cmap=cmap, norm=norm,
                      linewidths=line_width, zorder=-3,
                      )
    lines.set_array(amps)  # sets the colors of the segments


def _format_axes(fig_cont, limits, ncol):
    ylabels = fig_cont.data.keys()
    for idx_plane, plane in enumerate(PLANES):
        ax = fig_cont.axes[idx_plane]
        if ncol < 1:
            ax.set_yticklabels([])
            ax.set_yticks([])
        else:
            ax.set_yticklabels(ylabels, fontdict={'fontsize': get_fontsize_as_float(matplotlib.rcParams[u'axes.labelsize']) * .5})
            ax.set_yticks(range(len(ylabels)))
        ax.set_xlabel(LABEL_X)
        ax.set_ylabel(LABEL_Y_WATERFALL.format(plane=plane.upper()))
        ax.set_xlim(limits.xlim)
        ax.set_ylim([-.5, len(ylabels) - .5])
        ax.tick_params(axis='x', which='both')



"""
Accelerator
-----------

This module provides high-level classes to define most functionality of ``model.accelerators``.
It contains entrypoint the parent `Accelerator` class as well as other support classes.
"""
import re
from pathlib import Path
from typing import List

import numpy
import pandas as pd
import tfs
from generic_parser.entrypoint_parser import EntryPointParameters

from omc3.model.constants import (
    ERROR_DEFFS_TXT,
    JOB_MODEL_MADX,
    MODIFIER_TAG,
    MODIFIERS_MADX,
    TWISS_AC_DAT,
    TWISS_ADT_DAT,
    TWISS_BEST_KNOWLEDGE_DAT,
    TWISS_DAT,
    TWISS_ELEMENTS_DAT,
)
from omc3.utils import logging_tools
from omc3.utils.iotools import PathOrStr

LOG = logging_tools.get_logger(__name__)
CURRENT_DIR = Path(__file__).parent


class AccExcitationMode:
    # it is very important that FREE = 0
    FREE, ACD, ADT = range(3)


DRIVEN_EXCITATIONS = dict(acd=AccExcitationMode.ACD, adt=AccExcitationMode.ADT)


class AccElementTypes:
    """Defines the strings for the element types ``BPMS``, ``MAGNETS`` and ``ARC_BPMS``."""

    BPMS = "bpm"
    MAGNETS = "magnet"
    ARC_BPMS = "arc_bpm"


class Accelerator:
    """
    Abstract class to serve as an interface to implement the rest of the accelerators.
    """
    # RE_DICT needs to use MAD-X compatible patterns (jdilly, 2021)
    RE_DICT = {
        AccElementTypes.BPMS: r".*",
        AccElementTypes.MAGNETS: r".*",
        AccElementTypes.ARC_BPMS: r".*",
    }
    BPM_INITIAL = "B"

    @staticmethod
    def get_parameters():
        params = EntryPointParameters()
        params.add_parameter(
            name="model_dir",
            type=PathOrStr,
            help="Path to model directory; loads tunes and excitation from model!",
        )
        params.add_parameter(
            name="nat_tunes",
            type=float,
            nargs=2,
            help="Natural tunes without integer part.",
        )
        params.add_parameter(
            name="drv_tunes",
            type=float,
            nargs=2,
            help="Driven tunes without integer part.",
        )
        params.add_parameter(
            name="driven_excitation",
            type=str,
            choices=("acd", "adt"),
            help="Denotes driven excitation by AC-dipole (acd) or by ADT (adt)",
        )
        params.add_parameter(
            name="dpp",
            default=0.0,
            type=float,
            help="Delta p/p to use.",
        )
        params.add_parameter(
            name="energy",
            type=float,
            help="Energy in Tev.",
        )
        params.add_parameter(
            name="modifiers",
            type=Path,
            nargs="*",
            help="Path to the optics file to use (modifiers file).",
        )
        params.add_parameter(
            name="xing", action="store_true", help="If True, x-ing angles will be applied to model"
        )
        return params

    def __init__(self, opt) -> None:
        self.model_dir = None
        self.nat_tunes = None
        self.drv_tunes = None
        self.excitation = AccExcitationMode.FREE
        self.model = None
        self._model_driven = None
        self.model_best_knowledge = None
        self.elements = None
        self.error_defs_file = None
        self.modifiers = None
        self._beam_direction = 1
        self._beam = None
        self._ring = None
        self.energy = None
        self.dpp = 0.0
        self.xing = None

        if opt.model_dir:
            if (opt.nat_tunes is not None) or (opt.drv_tunes is not None):
                raise AcceleratorDefinitionError(
                    "Arguments 'nat_tunes' and 'driven_tunes' are "
                    "not allowed when loading from model directory."
                )
            self.init_from_model_dir(Path(opt.model_dir))

        else:
            self.init_from_options(opt)

    def init_from_options(self, opt) -> None:
        if opt.nat_tunes is None:
            raise AcceleratorDefinitionError("Argument 'nat_tunes' is required.")
        if (opt.drv_tunes is None) and (opt.driven_excitation is not None):
            raise AcceleratorDefinitionError("Argument 'drv_tunes' is required.")
        self.nat_tunes = opt.nat_tunes

        if opt.driven_excitation is not None:
            self.drv_tunes = opt.drv_tunes
            self.excitation = DRIVEN_EXCITATIONS[opt.driven_excitation]

        # optional with default
        self.dpp = opt.dpp

        # optional no default
        self.energy = opt.get("energy", None)
        self.xing = opt.get("xing", None)
        self.modifiers = opt.get("modifiers", None)

    def init_from_model_dir(self, model_dir: Path) -> None:
        LOG.debug("Creating accelerator instance from model dir")
        self.model_dir = model_dir
        # Elements #####################################
        elements_path = model_dir / TWISS_ELEMENTS_DAT
        if not elements_path.is_file():
            raise AcceleratorDefinitionError("Elements twiss not found")
        self.elements = tfs.read(elements_path, index="NAME")

        LOG.debug(f"  model path = {model_dir / TWISS_DAT}")
        try:
            self.model = tfs.read(model_dir / TWISS_DAT, index="NAME")
        except IOError:
            bpm_index = [
                idx for idx in self.elements.index.to_numpy() if idx.startswith(self.BPM_INITIAL)
            ]
            self.model = self.elements.loc[bpm_index, :]
        self.nat_tunes = [float(self.model.headers["Q1"]), float(self.model.headers["Q2"])]
        # self.energy = float(self.model.headers["ENERGY"]) * 1e-3  # TODO not the same Energy (jdilly, 2021)

        # Excitations #####################################
        driven_filenames = dict(acd=model_dir / TWISS_AC_DAT, adt=model_dir / TWISS_ADT_DAT)
        if driven_filenames["acd"].is_file() and driven_filenames["adt"].is_file():
            raise AcceleratorDefinitionError("ADT as well as ACD models provided. Choose only one.")
        for key in driven_filenames.keys():
            if driven_filenames[key].is_file():
                self._model_driven = tfs.read(driven_filenames[key], index="NAME")
                self.excitation = DRIVEN_EXCITATIONS[key]

        if not self.excitation == AccExcitationMode.FREE:
            self.drv_tunes = [self.model_driven.headers["Q1"], self.model_driven.headers["Q2"]]

        # Best Knowledge #####################################
        best_knowledge_path = model_dir / TWISS_BEST_KNOWLEDGE_DAT
        if best_knowledge_path.is_file():
            self.model_best_knowledge = tfs.read(best_knowledge_path, index="NAME")

        # Modifiers #########################################
        self.modifiers = _get_modifiers_from_modeldir(model_dir)

        # Error Def #####################################
        errordefspath = self.model_dir / ERROR_DEFFS_TXT
        if errordefspath.is_file():
            self.error_defs_file = errordefspath

    # Class methods ###########################################

    @classmethod
    def get_element_types_mask(cls, list_of_elements: List[str], types) -> numpy.ndarray:
        """
        Returns a boolean mask for elements in ``list_of_elements`` that belong to any of the
        specified types.
        Needs to handle: `bpm`, `magnet`, `arc_bpm` (see :class:`AccElementTypes`)

        Args:
            list_of_elements: list of elements.
            types: the kinds of elements to look for.

        Returns:
            A boolean array of elements of specified kinds.
        """
        unknown_elements = [ty for ty in types if ty not in cls.RE_DICT]
        if len(unknown_elements):
            raise TypeError(f"Unknown element(s): '{unknown_elements}'")
        series = pd.Series(list_of_elements)
        mask = series.str.match(cls.RE_DICT[types[0]], case=False)
        for ty in types[1:]:
            mask = mask | series.str.match(cls.RE_DICT[ty], case=False)
        return mask.to_numpy()

    @classmethod
    def get_variables(cls, frm=None, to=None, classes=None):
        """
        Gets the variables with elements in the given range and the given classes. ``None`` means
        everything.
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    @classmethod
    def get_correctors_variables(cls, frm=None, to=None, classes=None):
        """
        Returns the set of corrector variables between ``frm`` and ``to``, with classes in
        classes. ``None`` means select all.
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    @property
    def beam_direction(self) -> int:
        return self._beam_direction

    @beam_direction.setter
    def beam_direction(self, value: int) -> None:
        if value not in (1, -1):
            raise AcceleratorDefinitionError("Beam direction has to be either 1 or -1")
        self._beam_direction = value

    def verify_object(self):
        """
        Verifies that this instance of an `Accelerator` is properly instantiated.
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    def get_exciter_bpm(self, plane, distance):
        """
        Returns the BPM next to the exciter.
        The `Accelerator` instance knows already which excitation method is used.

        Args:
            plane: **X** or **Y**.
            distance: 1=nearest bpm 2=next to nearest bpm.
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    def important_phase_advances(self):
        return []

    @property
    def model_driven(self) -> tfs.TfsDataFrame:
        if self._model_driven is None:
            raise AttributeError("No driven model given in this accelerator instance.")
        return self._model_driven

    @classmethod
    def get_dir(cls) -> Path:
        """Default directory for accelerator. Should be overwritten if more specific."""
        dir_path = CURRENT_DIR / cls.NAME
        if dir_path.exists():
            return dir_path
        raise NotImplementedError(f"No directory for accelerator {cls.NAME} available.")

    @classmethod
    def get_file(cls, filename: str) -> Path:
        """Default location for accelerator files. Should be overwritten if more specific."""
        file_path = CURRENT_DIR / cls.NAME / filename
        if file_path.exists():
            return file_path
        raise NotImplementedError(
            f"File {file_path.name} not available for accelerator {cls.NAME}."
        )

    # Jobs ###################################################################

    def get_update_correction_script(self, tiwss_out_path, corrections_file_path):
        """
        Returns job (string) to create an updated model from changeparameters input (used in
        iterative correction).
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    def get_base_madx_script(self, best_knowledge=False):
        """
        Returns job (string) to create the basic accelerator sequence.
        """
        raise NotImplementedError("A function should have been overwritten, check stack trace.")

    ##########################################################################


class Variable:
    """
    Generic corrector variable class that holds `name`, `position (s)` and physical elements it
    affects. These variables should be logical variables that have and effect in the model if
    modified.
    """

    def __init__(self, name, elements, classes):
        self.name = name
        self.elements = elements
        self.classes = classes


class Element:
    """
    Generic corrector element class that holds `name` and `position (s)` of the corrector. This
    element should represent a physical element of the accelerator.
    """

    def __init__(self, name, s):
        self.name = name
        self.s = s


class AcceleratorDefinitionError(Exception):
    """
    Raised when an `Accelerator` instance is wrongly used, for example by calling a method that
    should have been overwritten.
    """

    pass


# Helper ----


def _get_modifiers_from_modeldir(model_dir: Path) -> List[Path]:
    """Parse modifiers from job.create_model.madx or use modifiers.madx file."""
    job_file = model_dir / JOB_MODEL_MADX
    if job_file.exists():
        job_madx = job_file.read_text()

        # find modifier tag in lines and return called file in these lines
        modifiers = re.findall(
            fr"\s+call,\s*file\s*=\s*[\"\']?([^;\'\"]+)[\"\']?\s*;\s*{MODIFIER_TAG}",
            job_madx,
            flags=re.IGNORECASE,
        )
        modifiers = [Path(m) for m in modifiers]
        return modifiers or None

    # Legacy
    modifiers_file = model_dir / MODIFIERS_MADX
    if modifiers_file.exists():  # legacy
        return [modifiers_file]

    return None



"""
Write MAD-X Macros
------------------

Write out madx scripts for the tracking macros.

**Arguments:**

*--Required--*

- **outputdir**:

    Directory where the observation_points.def will be put.

- **twissfile**:

    Path to twissfile with observationspoint in the NAME column.
"""

import tfs
from pathlib import Path
from omc3.model.constants import OBS_POINTS
from generic_parser.entrypoint_parser import entrypoint, EntryPointParameters

def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="twissfile",
        required=True,
        help="Path to twissfile with observationspoint in the NAME column."
        )
    params.add_parameter(
        name="outputdir",
        required=True,
        help=f"Directory where the {OBS_POINTS} will be put."
        )
    return params

@entrypoint(get_params(), strict=True)
def read_twiss_and_return_obs(opt):
    tracking_macros(tfs.read(opt['twissfile'], index='NAME').index.tolist(), Path(opt['outputdir']))

def _call(path_to_call):
    return f"call, file = '{path_to_call}';\n"


def define_observation_points_macros(list_of_bpms):
    macro = ""
    for tracker, prefix in (("ptc", "ptc_"), ("madx", "")):
        macro += f"define_{tracker}_observation_points(): macro = {{\n"
        macro += "".join([f"    {prefix}observe, place='{bpm}';\n" for bpm in list_of_bpms])
        macro += "};\n"
    return macro


def tracking_macros(list_of_bpms, outdir):
    obs_macro_file = outdir / OBS_POINTS
    with open(obs_macro_file, 'w') as obs_script:
        obs_script.write(define_observation_points_macros(list_of_bpms))
    track_macros = _call(obs_macro_file)
    track_macros += """
    /*
    * Performs a single particle tracking of the active sequence.
    * @param hkick: Horizontal kick magnitude in meters.
    * @param vkick: Vertical kick magnitude in meters.
    * @param num_turns: Number of turns to simulate.
    * @param output_path: Path to the output file, will add "one" at the end of the file name.
    */
    ptc_track_single_particle(hkick, vkick, num_turns, output_path): macro = {
    PTC_CREATE_UNIVERSE;
    PTC_CREATE_LAYOUT, model = 3, method = 6, nst = 10;
    PTC_ALIGN;
    EXEC, define_ptc_observation_points();
    PTC_START, x = hkick, y = vkick; ! kick units meters
    PTC_TRACK, deltap = 0.0, icase = 5, turns = num_turns, ELEMENT_BY_ELEMENT, dump, onetable, file = output_path;
    PTC_TRACK_END;
    PTC_END;
    }
    
    /*
    * Performs a single particle tracking of the active sequence using MAD-X track.
    * @param start_x: Particle horizontal start position.
    * @param start_y: Particle vertical start position.
    * @param start_px: Particle horizontal start momentum.
    * @param start_py: Particle vertical start momentum.
    * @param num_turns: Number of turns to simulate.
    * @param output_path: Path to the output file, will add "one" at the end of the file name.
    */
    madx_track_single_particle(start_x, start_y, start_px, start_py, num_turns, output_path): macro = {
    TRACK, DELTAP=0.0, ONETABLE, DUMP, FILE=output_path;
    START, X=start_x, PX=start_px, Y=start_y, PY=start_py;
    EXEC, define_madx_observation_points();
    RUN, TURNS=num_turns;
    ENDTRACK;
    }
    """
    return track_macros


if __name__ == "__main__":
    read_twiss_and_return_obs()



"""
Handler
-------

This module contains high-level functions to manage most functionality of the corrections calculations.
"""
import datetime
import os
import time
from pathlib import Path
from typing import Callable, Dict, List, Sequence, Tuple

import numpy as np
import pandas as pd
import tfs
from generic_parser import DotDict
from sklearn.linear_model import OrthogonalMatchingPursuit

import omc3.madx_wrapper as madx_wrapper
from omc3.correction import filters, model_appenders, response_twiss
from omc3.correction.constants import (BETA, DELTA, DIFF, DISP, ERROR, F1001,
                                       F1010, NORM_DISP, PHASE, TUNE,
                                       VALUE, WEIGHT)
from omc3.correction.model_appenders import add_coupling_to_model
from omc3.correction.response_io import read_fullresponse
from omc3.model.accelerators.accelerator import Accelerator
from omc3.optics_measurements.constants import (DISPERSION_NAME, EXT, REAL, IMAG,
                                                NORM_DISP_NAME, PHASE_NAME, NAME)
from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)


def correct(accel_inst: Accelerator, opt: DotDict) -> None:
    """ Perform global correction as described in :mod:`omc3.global_correction`.

    Args:
        accel_inst (Accelerator): Accelerator Instance
        opt (DotDict): Correction options,
                       see :mod:`omc3.global_correction` for details.

    """
    method_options = opt.get_subdict(["svd_cut", "n_correctors"])
    # read data from files
    vars_list = _get_varlist(accel_inst, opt.variable_categories)
    optics_params, meas_dict = get_measurement_data(
        opt.optics_params,
        opt.meas_dir,
        opt.beta_file_name,
        opt.weights,
    )

    if opt.fullresponse_path is not None:
        resp_dict = _load_fullresponse(opt.fullresponse_path, vars_list)
    else:
        resp_dict = response_twiss.create_response(accel_inst, opt.variable_categories, optics_params)

    # the model in accel_inst is modified later, so save nominal model here to variables
    nominal_model = _maybe_add_coupling_to_model(accel_inst.model, optics_params)
    # apply filters to data
    meas_dict = filters.filter_measurement(optics_params, meas_dict, nominal_model, opt)
    meas_dict = model_appenders.add_differences_to_model_to_measurements(nominal_model, meas_dict)

    resp_dict = filters.filter_response_index(resp_dict, meas_dict, optics_params)
    resp_matrix = _join_responses(resp_dict, optics_params, vars_list)
    delta = tfs.TfsDataFrame(0., index=vars_list, columns=[DELTA])

    # ######### Iteration Phase ######### #
    for iteration in range(opt.max_iter + 1):
        LOG.info(f"Correction Iteration {iteration} of {opt.max_iter}.")

        # ######### Update Model and Response ######### #
        if iteration > 0:
            LOG.debug("Updating model via MADX.")
            corr_model_path = opt.output_dir / f"twiss_{iteration}{EXT}"

            corr_model_elements = _create_corrected_model(corr_model_path, opt.change_params_path, accel_inst)
            corr_model_elements = _maybe_add_coupling_to_model(corr_model_elements, optics_params)

            bpms_index_mask = accel_inst.get_element_types_mask(corr_model_elements.index, types=["bpm"])
            corr_model = corr_model_elements.loc[bpms_index_mask, :]

            meas_dict = model_appenders.add_differences_to_model_to_measurements(corr_model, meas_dict)

            if opt.update_response:
                LOG.debug("Updating response.")
                # please look away for the next two lines.
                accel_inst._model = corr_model
                accel_inst._elements = corr_model_elements
                resp_dict = response_twiss.create_response(accel_inst, opt.variable_categories, optics_params)
                resp_dict = filters.filter_response_index(resp_dict, meas_dict, optics_params)
                resp_matrix = _join_responses(resp_dict, optics_params, vars_list)

        # ######### Actual optimization ######### #
        delta += _calculate_delta(resp_matrix, meas_dict, optics_params, vars_list, opt.method, method_options)

        # remove unused correctors from vars_list
        delta, resp_matrix, vars_list = _filter_by_strength(delta, resp_matrix, opt.min_corrector_strength)

        writeparams(opt.change_params_path, delta)
        writeparams(opt.change_params_correct_path, -delta)
        LOG.debug(f"Cumulative delta: {np.sum(np.abs(delta.loc[:, DELTA].to_numpy())):.5e}")
    write_knob(opt.knob_path, delta)
    LOG.info("Finished Iterative Global Correction.")


# Input ------------------------------------------------------------------------


def read_measurement_file(meas_dir: Path, filename: str) -> tfs.TfsDataFrame:
    return tfs.read(meas_dir / filename, index="NAME")


def get_measurement_data(
        keys: Sequence[str],
        meas_dir: Path,
        beta_file_name: str,
        w_dict: Dict[str, float] = None,
) -> Tuple[List[str], Dict[str, tfs.TfsDataFrame]]:
    """ Loads all measurements defined by `keys` into a dictionary. """
    measurement = {}
    filtered_keys = keys
    if w_dict is not None:
        filtered_keys = [key for key in keys if w_dict[key] != 0]

    for key in filtered_keys:
        if key.startswith(f"{PHASE}"):
            measurement[key] = read_measurement_file(meas_dir, f"{PHASE_NAME}{key[-1].lower()}{EXT}")

        elif key.startswith(f"{DISP}"):
            measurement[key] = read_measurement_file(meas_dir, f"{DISPERSION_NAME}{key[-1].lower()}{EXT}")

        elif key == f"{NORM_DISP}X":
            measurement[key] = read_measurement_file(meas_dir, f"{NORM_DISP_NAME}{key[-1].lower()}{EXT}")

        elif key[:5] in (F1010, F1001):
            measurement[key] = read_measurement_file(meas_dir, f"{key[:5].lower()}{EXT}")

        elif key == f"{TUNE}":
            measurement[key] = pd.DataFrame(
                {  # Just fractional tunes:
                    VALUE: np.remainder([read_measurement_file(meas_dir, f"{PHASE_NAME}x{EXT}")[f"{TUNE}1"],
                                         read_measurement_file(meas_dir, f"{PHASE_NAME}x{EXT}")[f"{TUNE}2"]],
                                        [1, 1]),
                    ERROR: np.array([0.001, 0.001])  # TODO measured errors not in the file
                },
                index=[f"{TUNE}1", f"{TUNE}2"],
            )

        elif key.startswith(f"{BETA}"):
            measurement[key] = read_measurement_file(meas_dir, f"{beta_file_name}{key[-1].lower()}{EXT}")
    return filtered_keys, measurement


def _load_fullresponse(full_response_path: Path, variables: Sequence[str]) -> dict:
    """
    Full response is dictionary of optics-parameter gradients upon
    a change of a single quadrupole strength
    """
    LOG.debug("Starting loading Full Response optics")
    full_response_data = read_fullresponse(full_response_path)

    # There is a check in read_fullresponse but there all variables need to be present.
    # Here only some. So I leave it like that (jdilly 2021-06-03)
    loaded_variables = [var for resp in full_response_data.values() for var in resp]
    if not any([var in loaded_variables for var in variables]):
        raise ValueError(
            "None of the given variables found in response matrix. Are you using the right categories?"
        )
    return full_response_data


# Data handling ----------------------------------------------------------------


def _get_varlist(accel_cls: Accelerator, variables: Sequence[str]):  # TODO: Virtual?
    varlist = np.array(accel_cls.get_variables(classes=variables))
    if len(varlist) == 0:
        raise ValueError("No variables found! Make sure your categories are valid!")
    return varlist


def _maybe_add_coupling_to_model(model: tfs.TfsDataFrame, keys: Sequence[str]) -> tfs.TfsDataFrame:
    """ Add coupling to the model, if terms corresponding to coupling RDTs are
    found in the provided keys.

    Args:
        model (tfs.TfsDataFrame): Twiss dataframe.
        keys (Sequence[str]):

    Returns:
        A TfsDataFrame with the added columns.
    """
    if any([key for key in keys if key.startswith("F1")]):
        return add_coupling_to_model(model)
    return model


def _create_corrected_model(twiss_out: str, change_params, accel_inst: Accelerator) -> tfs.TfsDataFrame:
    """ Use the calculated deltas in changeparameters.madx to create a corrected model """
    madx_script: str = accel_inst.get_update_correction_script(twiss_out, change_params)
    twiss_out_path = Path(twiss_out)
    madx_script = f"! Based on model '{accel_inst.model_dir}'\n" + madx_script
    madx_wrapper.run_string(
        madx_script,
        output_file=twiss_out_path.parent / f"job.create_{twiss_out_path.stem}.madx",
        log_file=twiss_out_path.parent / f"job.create_{twiss_out_path.stem}.log",
        cwd=accel_inst.model_dir,  # models are always run from there
    )
    return tfs.read(twiss_out, index=NAME)


def _join_responses(resp, keys, varslist):
    """ Returns matrix #BPMs * #Parameters x #variables """
    return (
        pd.concat(
            [resp[k] for k in keys],  # dataframes
            axis="index",  # axis to join along
            join="outer",  # =[pd.Index(varslist)]
            # other axes to use (pd Index obj required)
        ).reindex(columns=varslist).fillna(0.0)
    )


def _join_columns(col, meas, keys):
    """ Retuns vector: N= #BPMs * #Parameters (BBX, MUX etc.) """
    return np.concatenate([meas[key].loc[:, col].to_numpy() for key in keys], axis=0)


def _filter_by_strength(delta: pd.DataFrame, resp_matrix: pd.DataFrame, min_strength: float = 0):
    """ Remove too small correctors """
    delta = delta.loc[delta[DELTA].abs() > min_strength]
    return delta, resp_matrix.loc[:, delta.index], delta.index.to_numpy()


# Optimization -----------------------------------------------------------------


def _get_method_fun(method: str) -> Callable:
    method_to_function_dict = {"pinv": _pseudo_inverse, "omp": _orthogonal_matching_pursuit}
    return method_to_function_dict[method]


def _pseudo_inverse(response_mat: pd.DataFrame, diff_vec, opt: DotDict):
    """ Calculates the pseudo-inverse of the response via svd. (numpy) """
    if opt.svd_cut is None:
        raise ValueError("svd_cut setting needed for pseudo inverse method.")
    return np.dot(np.linalg.pinv(response_mat, opt.svd_cut), diff_vec)


def _orthogonal_matching_pursuit(response_mat: pd.DataFrame, diff_vec, opt: DotDict):
    """ Calculated n_correctors via orthogonal matching pursuit"""
    if opt.n_correctors is None:
        raise ValueError("n_correctors setting needed for orthogonal matching pursuit.")

    res = OrthogonalMatchingPursuit(opt.n_correctors).fit(response_mat, diff_vec)
    coef = res.coef_
    LOG.debug(f"Orthogonal Matching Pursuit Results: \n"
              f"  Chosen variables: {response_mat.columns.to_numpy()[coef.nonzero()]}\n"
              f"  Score: {res.score(response_mat, diff_vec)}")
    return coef


def _calculate_delta(
        resp_matrix: pd.DataFrame,
        meas_dict: dict,
        keys: Sequence[str],
        vars_list: Sequence[str],
        method: str,
        meth_opt
):
    """Get the deltas for the variables.

    Output is Dataframe with one column 'DELTA' and vars_list index."""
    weight_vector = _join_columns(f"{WEIGHT}", meas_dict, keys)
    diff_vector = _join_columns(f"{DIFF}", meas_dict, keys)

    resp_weighted = resp_matrix.mul(weight_vector, axis="index")
    diff_weighted = diff_vector * weight_vector

    delta = _get_method_fun(method)(resp_weighted, diff_weighted, meth_opt)
    delta = tfs.TfsDataFrame(delta, index=vars_list, columns=[DELTA])

    # check calculations
    update = np.dot(resp_weighted, delta[DELTA])
    _print_rms(meas_dict, diff_weighted, update)

    return delta


# Print ------------------------------------------------------------------------


def _print_rms(meas: dict, diff_w, r_delta_w) -> None:
    """ Prints current RMS status """
    f_str = "{:>20s} : {:.5e}"
    LOG.debug("RMS Measure - Model (before correction, w/o weigths):")
    for key in meas:
        LOG.debug(f_str.format(key, _rms(meas[key].loc[:, DIFF].to_numpy())))

    LOG.info("RMS Measure - Model (before correction, w/ weigths):")
    for key in meas:
        LOG.info(f_str.format(key, _rms(meas[key].loc[:, DIFF].to_numpy() * meas[key].loc[:, WEIGHT].to_numpy())))

    LOG.info(f_str.format("All", _rms(diff_w)))
    LOG.debug(f_str.format("R * delta", _rms(r_delta_w)))
    LOG.debug("(Measure - Model) - (R * delta)   ")
    LOG.debug(f_str.format("", _rms(diff_w - r_delta_w)))


def _rms(a):
    """Calculate root mean square error of a number or an ndarray."""
    return np.sqrt(np.mean(np.square(a)))


# Output -----------------------------------------------------------------------


def write_knob(knob_path: Path, delta: pd.DataFrame) -> None:
    a = datetime.datetime.fromtimestamp(time.time())
    delta_out = -delta.loc[:, [DELTA]]
    delta_out.headers["PATH"] = str(knob_path.parent)
    delta_out.headers["DATE"] = str(a.ctime())
    tfs.write(knob_path, delta_out, save_index="NAME")


def writeparams(path_to_file: Path, delta: pd.DataFrame) -> None:
    with open(path_to_file, "w") as madx_script:
        for var in delta.index.to_numpy():
            value = delta.loc[var, DELTA]
            madx_script.write(f"{var} = {var} {value:+e};\n")



"""
Constants
---------

Constants and definitions for the ``plotting`` module.
"""
from omc3.optics_measurements.constants import (
    S, ERR, DELTA, AMPLITUDE, BETA, PHASE, REAL, IMAG,
    AMP_BETA_NAME, BETA_NAME, CHROM_BETA_NAME, PHASE_NAME,
    SPECIAL_PHASE_NAME, TOTAL_PHASE_NAME, DISPERSION_NAME,
    NORM_DISP_NAME, ORBIT_NAME, KICK_NAME, IP_NAME, PHASE_ADV, MDL
)
from omc3.plotting.utils.annotations import ylabels

IP_POS_DEFAULT = {
    "LHCB1": {
        'IP1': 23519.36962,
        'IP2': 192.923,
        'IP3': 3525.207216,
        'IP4': 6857.491433,
        'IP5': 10189.77565,
        'IP6': 13522.21223,
        'IP7': 16854.64882,
        'IP8': 20175.8654,
    },
    "LHCB2": {
        'IP1': 3195.252584,
        'IP2': 6527.5368,
        'IP3': 9859.973384,
        'IP4': 13192.40997,
        'IP5': 16524.84655,
        'IP6': 19857.13077,
        'IP7': 23189.41498,
        'IP8': 26510.4792,
    }
}

MANUAL_STYLE = {
    # differences to the standard style
    u'lines.markersize': 5.0,
    u'lines.linestyle': u'',
    # u'figure.figsize': [6, 3.8],
}

COMPLEX_NAMES = [p+ext for p in ["1001", "1010"] for ext in "RI"]  # Endings of columns that contain complex data

DEFAULTS = {
    'ncol_legend': 3,
    'errorbar_alpha': .6,
}


XAXIS = { # Column, Label
    'location': (S, 'Location [m]'),
    'phase-advance': (f'{PHASE_ADV}{{0}}{MDL}', 'Phase Advance [$2 \pi$]')
}

YAXIS = {  # column, column label, yaxis-label [, Delta yaxis-label]
    BETA_NAME:        (BETA, 'beta', ylabels['beta'], ylabels['betabeat']),
    AMP_BETA_NAME:    (BETA, 'beta', ylabels['beta'], ylabels['betabeat']),
    ORBIT_NAME:       ('', 'orbit', ylabels['co'], f"$\Delta$ {ylabels['co']}"),
    PHASE_NAME:       (PHASE, 'phase', ylabels['phase']),
    TOTAL_PHASE_NAME: (PHASE, 'total phase', ylabels['phase']),
    'rdt_amp':        (AMPLITUDE, None, ylabels['absolute']),
    'rdt_phase':      (PHASE, None, ylabels['phase']),
    'rdt_real':       (REAL, None, ylabels['real']),
    'rdt_imag':       (IMAG, None, ylabels['imag']),
}



"""
Utils
-----

This module contains utilities for the ``plotting`` module.
"""
from collections import OrderedDict
from pathlib import Path
from typing import Iterable

from matplotlib import pyplot as plt


class FigureContainer:
    """Container for attaching additional information to one figure."""
    def __init__(self, id_: str, path: Path, axes_ids: Iterable[str]) -> None:
        self.fig, axs = plt.subplots(nrows=len(axes_ids))
        self.fig.canvas.manager.set_window_title(id_)

        if len(axes_ids) == 1:
            axs = [axs]

        self.axes_ids = axes_ids  # to keep order
        self.axes = {ax_id: ax for ax_id, ax in zip(axes_ids, axs)}
        self.xlabels = {ax_id: None for ax_id in axes_ids}
        self.ylabels = {ax_id: None for ax_id in axes_ids}
        self.data = {ax_id: {} for ax_id in axes_ids}
        self.path = path

    def __getitem__(self, ax_id):
        return self.axes[ax_id], self.data[ax_id], self.xlabels[ax_id], self.ylabels[ax_id]


class FigureCollector:
    """Class to collect figure containers and manage data adding."""
    def __init__(self) -> None:
        self.fig_dict = OrderedDict()   # dictionary of matplotlib figures, for output
        self.figs = OrderedDict()       # dictionary of FigureContainers, used internally

    def add_data_for_id(self, figure_id: str, label: str, data: dict,
                        x_label: str, y_label: str,
                        path: Path = None, axes_id: str = '',
                        axes_ids: Iterable[str] = ('',)) -> None:
        """Add the data at the appropriate figure container."""
        try:
            figure_cont = self.figs[figure_id]
        except KeyError:
            figure_cont = FigureContainer(figure_id, path, axes_ids)

            self.figs[figure_id] = figure_cont
            self.fig_dict[figure_id] = figure_cont.fig

        figure_cont.ylabels[axes_id] = y_label
        figure_cont.xlabels[axes_id] = x_label
        figure_cont.data[axes_id][label] = data



"""
Analysis
--------

This module contains the analysis functionality of ``kmod``.
It provides functions to calculate beta functions at different locations from K-modulation data.
"""
import datetime

import numpy as np
import scipy.optimize
import tfs
from tfs import tools as tfstools

from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.kmod import helper
from omc3.kmod.constants import CLEANED, K, TUNE, ERR, BETA, STAR, WAIST, PHASEADV, AVERAGE, \
    SEQUENCES_PATH
from omc3.model.constants import TWISS_DAT
from omc3.optics_measurements.constants import PHASE_NAME, EXT
from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)


def return_sign_for_err(n):
    """
    Creates an array for error calculation, of form:
    [[ 0.  0.  0.]
    [ 1.  0.  0.]
    [-1. -0. -0.]
    [ 0.  1.  0.]
    [-0. -1. -0.]
    [ 0.  0.  1.]
    [-0. -0. -1.]]
    Columns corresponds to error, i.e. first column for `dQ` etc.
    """
    sign = np.zeros((2*n+1, n))

    sign[1::2] = np.eye(n)
    sign[2::2] = -np.eye(n)
    return sign


def propagate_beta_in_drift(beta_waist, drift):
    beta = beta_waist + drift**2/beta_waist
    return beta


def calc_betastar(kmod_input_params, results_df, l_star):

    sign = return_sign_for_err(2)      

    for plane in PLANES:
        betastar = propagate_beta_in_drift((results_df.loc[:, f"{BETA}{WAIST}{plane}"].to_numpy() + sign[:, 0] * results_df.loc[:, f"{ERR}{BETA}{WAIST}{plane}"].to_numpy()),
                                           (results_df.loc[:, f"{WAIST}{plane}"].to_numpy() + sign[:, 1] * results_df.loc[:, f"{ERR}{WAIST}{plane}"].to_numpy()))
        betastar_err = get_err(betastar[1::2]-betastar[0])

        if kmod_input_params.no_sig_digits:
            results_df[f"{BETA}{STAR}{plane}"], results_df[f"{ERR}{BETA}{STAR}{plane}"] = (betastar[0], betastar_err)
        else:
            results_df[f"{BETA}{STAR}{plane}"], results_df[f"{ERR}{BETA}{STAR}{plane}"] = tfstools.significant_digits(betastar[0], betastar_err, return_floats=True)


    # reindex df to put betastar first
    cols = results_df.columns.tolist()
    cols = [cols[0]]+cols[-4:]+cols[1:-4]
    results_df = results_df.reindex(columns=cols)

    for plane in PLANES:
        results_df[f"{PHASEADV}{plane}"], results_df[f"{ERR}{PHASEADV}{plane}"] = phase_adv_from_kmod(
            l_star, betastar[0], betastar_err,
            results_df.loc[:, f"{WAIST}{plane}"].to_numpy(),
            results_df.loc[:, f"{ERR}{WAIST}{plane}"].to_numpy())

    return results_df


def phase_adv_from_kmod(lstar, betastar, ebetastar, waist, ewaist):
    return _phase_adv_from_kmod_value(lstar, betastar, waist),\
           _phase_adv_from_kmod_err(lstar, betastar, ebetastar, waist, ewaist)


def _phase_adv_from_kmod_value(lstar, betastar, waist):
    return (np.arctan((lstar - waist) / betastar) +
            np.arctan((lstar + waist) / betastar)) / (2 * np.pi)


def _phase_adv_from_kmod_err(lstar, betastar, ebetastar, waist, ewaist):
    numer = (2 * lstar * (betastar ** 2 + lstar ** 2 - waist ** 2) * ebetastar) ** 2
    numer = numer + (4 * betastar * lstar * waist * ewaist) ** 2
    denom = (betastar ** 2 + (lstar - waist) ** 2) ** 2
    denom = denom * (betastar ** 2 + (lstar + waist) ** 2) ** 2
    return np.sqrt(numer / denom) / (2 * np.pi)


def calc_beta_inst(name, position, results_df, magnet1_df, magnet2_df, kmod_input_params):
    betas = np.zeros((2, 2))
    sign = np.array([[0, 0], [1, 0], [-1, 0], [0, 1], [0, -1]])
    for i, plane in enumerate(PLANES):
        waist = results_df.loc[:, f"{WAIST}{plane}"].to_numpy()
        if magnet1_df.headers['POLARITY'] == 1 and magnet2_df.headers['POLARITY'] == -1:
            waist = -waist
        if plane == 'Y':
            waist = -waist

        beta = propagate_beta_in_drift((results_df.loc[:, f"{BETA}{WAIST}{plane}"].to_numpy() + sign[:, 0] * results_df.loc[:, f"{ERR}{BETA}{WAIST}{plane}"].to_numpy()),
                                       ((waist - position) + sign[:, 1] * results_df.loc[:, f"{ERR}{WAIST}{plane}"].to_numpy()))
        beta_err = get_err(beta[1::2]-beta[0])
        if kmod_input_params.no_sig_digits:
            betas[i, 0], betas[i, 1] = beta[0], beta_err
        else:
            betas[i, 0], betas[i, 1] = tfstools.significant_digits(beta[0], beta_err, return_floats=True)

    return name, betas[0, 0], betas[0, 1], betas[1, 0], betas[1, 1]


def calc_beta_at_instruments(kmod_input_params, results_df, magnet1_df, magnet2_df):

    beta_instr = []

    for instrument in kmod_input_params.instruments_found:
        positions = getattr(kmod_input_params, instrument)

        for name, position in positions.items():
            beta_instr.append(calc_beta_inst(
                name, position, results_df, magnet1_df, magnet2_df, kmod_input_params))

    instrument_beta_df = tfs.TfsDataFrame(
        columns=['NAME',
                 f"{BETA}{'X'}",
                 f"{ERR}{BETA}{'X'}",
                 f"{BETA}{'Y'}",
                 f"{ERR}{BETA}{'Y'}",
                 ],
        data=beta_instr)

    return instrument_beta_df


def fit_prec(x, beta_av):
    twopiQ = 2 * np.pi * np.modf(x[1])[0]
    dQ = (1/(2.*np.pi)) * np.arccos(np.cos(twopiQ) -
                                    0.5 * beta_av * x[0] * np.sin(twopiQ)) - np.modf(x[1])[0]
    return dQ


np.vectorize(fit_prec)


def fit_approx(x, beta_av):
    dQ = beta_av*x[0]/(4*np.pi)
    return dQ


np.vectorize(fit_approx)


def average_beta_from_Tune(Q, TdQ, l, Dk):
    """Calculates average beta function in quadrupole from tune change ``TdQ`` and ``delta K``."""

    beta_av = 2 * (1 / np.tan(2 * np.pi * Q) *
              (1 - np.cos(2 * np.pi * TdQ)) + np.sin(2 * np.pi * TdQ)) / (l * Dk)
    return abs(beta_av)


def average_beta_focussing_quadrupole(b, w, L, K, Lstar):

    beta0 = b + ((Lstar - w) ** 2 / b)
    alpha0 = -(Lstar - w) / b
    average_beta = (beta0/2.) * (1 + ((np.sin(2 * np.sqrt(abs(K)) * L)) / (2 * np.sqrt(abs(K)) * L))) \
                   - alpha0 * ((np.sin(np.sqrt(abs(K)) * L)**2) / (abs(K) * L)) \
                   + (1/(2*abs(K))) * ((1 + alpha0**2) / beta0) * \
                   (1 - ((np.sin(2 * np.sqrt(abs(K)) * L)) / (2 * np.sqrt(abs(K)) * L)))

    return average_beta


np.vectorize(average_beta_focussing_quadrupole)


def average_beta_defocussing_quadrupole(b, w, L, K, Lstar):
    beta0 = b + ((Lstar - w) ** 2 / b)
    alpha0 = -(Lstar - w) / b
    average_beta = (beta0/2.) * (1 + ((np.sinh(2 * np.sqrt(abs(K)) * L)) / (2 * np.sqrt(abs(K)) * L))) \
                   - alpha0 * ((np.sinh(np.sqrt(abs(K)) * L)**2) / (abs(K) * L)) \
                   + (1/(2*abs(K))) * ((1 + alpha0**2) / beta0) * \
                   (((np.sinh(2 * np.sqrt(abs(K)) * L)) / (2 * np.sqrt(abs(K)) * L)) - 1)

    return average_beta


np.vectorize(average_beta_defocussing_quadrupole)


def calc_tune(magnet_df):
    for plane in PLANES:
        magnet_df.headers[f"{TUNE}{plane}"] = np.average(magnet_df.where(
            magnet_df[f"{CLEANED}{plane}"])[f"{TUNE}{plane}"].dropna())
    return magnet_df


def calc_k(magnet_df):
    magnet_df.headers[K] = np.average(magnet_df.where(magnet_df[f"{CLEANED}X"])[K].dropna())
    return magnet_df


def return_fit_input(magnet_df, plane):

    x = np.zeros((2, len(magnet_df.where(magnet_df[f"{CLEANED}{plane}"])[K].dropna())))

    sign = magnet_df.headers['POLARITY'] if plane == 'X' else -1 * magnet_df.headers['POLARITY']
    x[0, :] = sign*(
            magnet_df.where(magnet_df[f"{CLEANED}{plane}"])[K].dropna() -
            magnet_df.headers[K]) * magnet_df.headers['LENGTH']
    x[1, :] = magnet_df.headers[f"{TUNE}{plane}"]

    return x


def do_fit(magnet_df, plane, use_approx=False):
    if not use_approx:
        fun = fit_prec
    elif use_approx:
        fun = fit_approx
    
    sigma = magnet_df.where(magnet_df[f"{CLEANED}{plane}"])[f"{ERR}{TUNE}{plane}"].dropna()
    if not np.any(sigma):
        sigma = 1.E-22 * np.ones(len(sigma))

    av_beta, av_beta_err = scipy.optimize.curve_fit(
        fun,
        xdata=return_fit_input(magnet_df, plane),
        ydata=magnet_df.where(magnet_df[f"{CLEANED}{plane}"])[
            f"{TUNE}{plane}"].dropna() - magnet_df.headers[f"{TUNE}{plane}"],
        sigma=sigma,
        absolute_sigma=True,
        p0=1
    )
    return np.abs(av_beta[0]), np.sqrt(np.diag(av_beta_err))[0]


def get_av_beta(magnet_df):
    for plane in PLANES:
        magnet_df.headers[f"{AVERAGE}{BETA}{plane}"], magnet_df.headers[f"{ERR}{AVERAGE}{BETA}{plane}"] = do_fit(magnet_df, plane)
    return magnet_df


def check_polarity(magnet1_df, magnet2_df, sign):
    left, right = sign
    return magnet1_df.headers['POLARITY'] == left and magnet2_df.headers['POLARITY'] == right


def return_df(magnet1_df, magnet2_df, plane):

    sign = {'X': np.array([1, -1]), 'Y': np.array([-1, 1])}

    if check_polarity(magnet1_df, magnet2_df, sign[plane]):
        return magnet1_df, magnet2_df
    elif check_polarity(magnet1_df, magnet2_df, -sign[plane]):
        return magnet2_df, magnet1_df

      
def get_BPM(kmod_input_params):

    # listing the BPMs of the last quadrupole BPMs
    BPM_dict = {
        "IP1": ["BPMSW.1L1", "BPMSW.1R1"],
        "IP2": ["BPMSW.1L2", "BPMSW.1R2"],
        "IP3": ["BPMW.4L3", "BPMW.4R3"],
        "IP4": ["BPMWA.A5L4", "BPMWA.A5R4"],
        "IP5": ["BPMSW.1L5", "BPMSW.1R5"],
        "IP6": ["BPMSE.4L6", "BPMSA.4R6"],
        "IP7": ["BPMW.4L7", "BPMW.4R7"],
        "IP8": ["BPMSW.1L8", "BPMSW.1R8"],
    }
    if kmod_input_params.interaction_point:
        return [f"{bpm}.B{kmod_input_params.beam:d}"
                for bpm in BPM_dict[kmod_input_params.interaction_point.upper()]]
    if kmod_input_params.circuits:
        return [f"{bpm}.B{kmod_input_params.beam:d}"
                for bpm in BPM_dict[f"IP{kmod_input_params.circuits[0][-3]}"]]
    raise AttributeError("Should not have happened, was checked in analyse_kmod")


def get_BPM_distance(kmod_input_params, BPML, BPMR):
    twiss_df = tfs.read(
        SEQUENCES_PATH / f"twiss_lhcb{kmod_input_params.beam:d}.dat", index='NAME'
    )
    return np.abs(twiss_df.loc[BPMR, 'S'] - twiss_df.loc[BPML, 'S']) / 2


def get_phase_from_model(kmod_input_params, plane):
    """Get the phase from twiss model."""
    twiss_df = tfs.read(kmod_input_params.model_dir / TWISS_DAT, index='NAME')
    BPML, BPMR = get_BPM(kmod_input_params)[0], get_BPM(kmod_input_params)[1]
    phase_adv_model = abs(twiss_df.loc[BPMR, f'MU{plane}'] - twiss_df.loc[BPML, f'MU{plane}'])
    phase_adv_err = 0.5e-3  # this number is given by Andrea's estimations

    return phase_adv_model, phase_adv_err


def get_phase_from_measurement(kmod_input_params, plane):
    phase_df = tfs.read(
        kmod_input_params.measurement_dir / f'{PHASE_NAME}{plane.lower()}{EXT}', index='NAME'
    )
    bpms_lr = get_BPM(kmod_input_params)
    for bpm in bpms_lr:
        if bpm not in phase_df.index.to_numpy():
            raise ValueError(f"BPM {bpm} not found in the measurement")
    # both BPMs are in measurement and there should be no other BPM in between
    return (phase_df.loc[bpms_lr[0], f'PHASE{plane}'],
            phase_df.loc[bpms_lr[0], f'{ERR}PHASE{plane}'])


def phase_constraint(kmod_input_params, plane):
    if kmod_input_params.measurement_dir:
        return get_phase_from_measurement(kmod_input_params, plane)
    # model is taken (if exists) in case no measurement data is provided
    if kmod_input_params.model_dir:
        return get_phase_from_model(kmod_input_params, plane)
    return [1.0, 1.0]


def chi2(x, foc_magnet_df, def_magnet_df, plane, kmod_input_params, sign, BPM_distance, phase_adv_constraint):

    b = x[0]
    w = x[1]

    if kmod_input_params.interaction_point:

        phase_adv = phase_adv_from_kmod(BPM_distance, b, 0.0, w, 0.0)[0]
        weight = kmod_input_params.phase_weight
    else:
        phase_adv = 0.0
        weight = 0

    c2 = (1-weight)*(((average_beta_focussing_quadrupole(b, w, foc_magnet_df.headers['LENGTH'] +
        sign[0] * kmod_input_params.errorL, foc_magnet_df.headers[K] +
        sign[1] * kmod_input_params.errorK * foc_magnet_df.headers[K],
        foc_magnet_df.headers['LSTAR'] +
        sign[2] * kmod_input_params.misalignment) -
        foc_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"] +
        sign[3] * foc_magnet_df.headers[f"{ERR}{AVERAGE}{BETA}{plane}"])/(def_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"] + foc_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"])/2.0) ** 2 +
        ((average_beta_defocussing_quadrupole(b, -w, def_magnet_df.headers['LENGTH'] +
        sign[4] * kmod_input_params.errorL, def_magnet_df.headers[K] +
        sign[5] * kmod_input_params.errorK * def_magnet_df.headers[K],
        def_magnet_df.headers['LSTAR'] +
        sign[6] * kmod_input_params.misalignment) -
        def_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"] +
        sign[7] * def_magnet_df.headers[f"{ERR}{AVERAGE}{BETA}{plane}"])/(foc_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"] + def_magnet_df.headers[f"{AVERAGE}{BETA}{plane}"])/2.0) ** 2) + \
        weight*(((phase_adv - (phase_adv_constraint[0]+sign[8]*phase_adv_constraint[1]))/phase_adv_constraint[0])**2)

    return c2


def get_beta_waist(magnet1_df, magnet2_df, kmod_input_params, plane):

    n = 9
    sign = return_sign_for_err(n)
    foc_magnet_df, def_magnet_df = return_df(magnet1_df, magnet2_df, plane)
    results = np.zeros((2*n+1, 2))
    BPML, BPMR = get_BPM(kmod_input_params)
    BPM_distance = get_BPM_distance(kmod_input_params, BPML, BPMR)
    phase_adv_constraint = phase_constraint(kmod_input_params, plane)
    for i, s in enumerate(sign):

        def fun(x): return chi2(x, foc_magnet_df, def_magnet_df, plane, kmod_input_params, s, BPM_distance, phase_adv_constraint)
        fitresults = scipy.optimize.minimize(fun=fun,
                                             x0=kmod_input_params.betastar_and_waist[plane],
                                             method='nelder-mead',
                                             tol=1E-22)

        results[i, :] = fitresults.x[0], fitresults.x[1]

    beta_waist_err = get_err(results[1::2, 0]-results[0, 0])
    waist_err = get_err(results[1::2, 1]-results[0, 1])

    return results[0, 0], beta_waist_err, results[0, 1], waist_err


def get_err(diff_array):
    return np.sqrt(np.sum(np.square(diff_array)))


def analyse(magnet1_df, magnet2_df, opt, betastar_required):

    for magnet_df in (magnet1_df, magnet2_df):
        LOG.info(f'Analysing magnet {magnet_df.headers["QUADRUPOLE"]}')
        magnet_df = helper.add_tune_uncertainty(magnet_df, opt.tune_uncertainty)
        magnet_df = helper.clean_data(magnet_df, opt.no_autoclean)
        magnet_df = calc_tune(magnet_df)
        magnet_df = calc_k(magnet_df)
        magnet_df = get_av_beta(magnet_df)

    LOG.info('Simplex to determine beta waist')
    results = {plane: get_beta_waist(magnet1_df, magnet2_df, opt, plane) for plane in PLANES}

    results_df = tfs.TfsDataFrame(
        columns=['LABEL',
                 "TIME"],
        data=[np.hstack((opt.label,
                         datetime.datetime.now().strftime(formats.TIME)))])

    for plane in PLANES:
        results_df[f"{BETA}{WAIST}{plane}"] = results[plane][0]
        results_df[f"{ERR}{BETA}{WAIST}{plane}"] = results[plane][1]
        results_df[f"{WAIST}{plane}"] = results[plane][2]
        results_df[f"{ERR}{WAIST}{plane}"] = results[plane][3]

    LOG.info('Calculate betastar')
    if betastar_required:
        results_df = calc_betastar(opt, results_df, magnet1_df.headers['LSTAR'])

    LOG.info('Calculate beta at instruments')
    if opt.instruments_found:
        instrument_beta_df = calc_beta_at_instruments(opt, results_df, magnet1_df, magnet2_df)


    return magnet1_df, magnet2_df, results_df, instrument_beta_df
